import sys, requests, json, time, threading, signal, subprocess, os, ConfigParser
import logger as logmodule
import cPickle as pickle

logger = logmodule.getLogger(__name__)

def load_config():
	parser = ConfigParser.ConfigParser(allow_no_value=True)
	parser.read(os.path.join(os.path.dirname(__file__),'conf/defaults.ini'))
	parser.read('/etc/c2pbbridge/c2pbridge.ini')
	
	return parser

def start():
	config = load_config()

	log_level = config.get('daemon', 'log_level').upper()

	if(log_level in logmodule.level_names):
		logmodule.logger_root.setLevel(logmodule.level_names[log_level])
	
	pid_file = config.get('daemon', 'pid_file')

	admin = BridgeAdmin(config.get('network', 'conf_url'), config.get('network', 'private_backbone'), config.get('network', 'vxlan_group'), config.get('network', 'vxlan_port', int))
	signal.signal(signal.SIGTERM, admin.signal)
	signal.signal(signal.SIGINT, admin.signal)
	signal.signal(signal.SIGHUP, admin.signal)
	admin.start()

	try:
		pid_file_h = open(pid_file, 'w')
		pid_file_h.write(str(os.getpid()))
		pid_file_h.close()
	
	except:
		admin.stop_event.set()
		sys.stderr.write('c2pbridge couldn\'t start...\n')
	

	while admin.isAlive():
		time.sleep(.1)

	os.unlink(pid_file)


class BridgeAdmin(threading.Thread):
	"""docstring for BridgeAdmin"""
	def __init__(self, conf_url, private_backbone, vxlan_group = '239.74.234.4', vxlan_port = 4789):
		super(BridgeAdmin, self).__init__()
		self.daemon = True
		self.bridge_data_list = None
		self.stop_event = threading.Event()
		self.update_event = threading.Event()
		self.active_ifaces = []
		self.bridges = {}
		self._lock = threading.Lock()
		self.conf_url = conf_url
		self.private_backbone_iface = private_backbone
		self.vxlan_group = vxlan_group
		self.vxlan_port = vxlan_port

	def call_wrapper(self, data):
		logger.debug(' '.join(data))
		return subprocess.check_call(data)
	
	
	def fetch_bridge_data(self):
		r = requests.get(self.conf_url)
		if r.status_code != 200:
			logger.error('fetch config failed (HTTP - {})'.format(r.status_code))
			return
		
		try:
			received_data = r.json()
			
		except:
			logger.error('unable to decode data')
		
		if pickle.dumps(self.bridge_data_list) != pickle.dumps(received_data):
			self.bridge_data_list = received_data
			self.apply_configuration()
			
	def iface_exists(self,iface):
		if os.path.exists('/sys/class/net/{}'.format(iface)):
			return True
			
		else:
			return False
	
	def create_vxlan(self,iface_data):
		assert iface_data['public'],'Should not create private vxlan interface'
		name = 'vp{}'.format(iface_data['vxlan_id'])
		assert not self.iface_exists(name),'Interface already exists'

		try:
			self.call_wrapper(['ip','link','add',name,
				'mtu','1500',
				'type','vxlan',
				'id','0x{}'.format(iface_data['vxlan_id']),
				'group',self.vxlan_group,
				'dev',self.private_backbone_iface,
				'dstport',str(self.vxlan_port)
			])
			self.call_wrapper(['ip','link','set','up','dev',name])
			return True
			
		except:
			return False
	
	def create_vlan(self,iface_data):
		if(iface_data['public']):
			name = 'vp{}'.format(iface_data['vxlan_id'])
			
		else:
			name = 'vv{}'.format(iface_data['vxlan_id'])
		
		assert self.iface_exists(name),'vxlan interface doesn\'t exists'
		
		try:
			self.call_wrapper(['ip', 'link', 'add', 'link', name, 
				name + '.{}'.format(iface_data['vlan_id']),
				'mtu', '1500',
				'type', 'vlan',
				'proto', '802.1Q', 
				'id', str(iface_data['vlan_id'])
			])
			self.call_wrapper(['ip','link','set','up','dev',name + '.{}'.format(iface_data['vlan_id'])])	
			return True
			
		except:
			return False
			
	def remove_iface(self, iface_name):
		assert self.iface_exists(iface_name),'Interface already exists'

		try:
			self.call_wrapper(['ip', 'link', 'set', 'down', 'dev', iface_name])
			self.call_wrapper(['ip', 'link', 'delete', iface_name])
			return True
			
		except:
			return False
			
	def create_bridge(self,bridge_name):
		assert not self.iface_exists(bridge_name),'Bridge already exists'
		
		try:
			self.call_wrapper(['brctl','addbr',bridge_name])
			self.call_wrapper(['brctl','stp',bridge_name,'on'])
			fh = open('/sys/devices/virtual/net/{}/bridge/multicast_querier'.format(bridge_name),'w')
			fh.write('1')
			fh.close()
			self.call_wrapper(['ip','link','set','up','dev',bridge_name])
			self.bridges[bridge_name] = []
			return True
		except:
			subprocess.call(['brctl','delbr',bridge_name])
			return False
			
	def remove_bridge(self, bridge_name):
		assert self.iface_exists(bridge_name), 'Interface already exists'
		assert len(self.bridges[bridge_name]) == 0, 'Bridge is not empty'
		try:
			self.call_wrapper(['ip', 'link', 'set', 'down', 'dev', bridge_name])
			self.call_wrapper(['brctl', 'delbr', bridge_name])
			del self.bridges[bridge_name]
			return True
			
		except:
			return False
	
	def add_bridge_member(self, bridge_name, iface_name):
		assert self.iface_exists(bridge_name),'Bridge doesn\'t exists'
		assert self.iface_exists(iface_name),'Interface doesn\'t exists'
		
		try:
			self.call_wrapper(['brctl','addif', bridge_name, iface_name])
			self.bridges[bridge_name].append(iface_name)
			return True
			
		except:
			return False
			
	def remove_bridge_member(self, bridge_name, iface_name):
		assert self.iface_exists(bridge_name),'Bridge doesn\'t exists'
		assert self.iface_exists(iface_name),'Interface doesn\'t exists'
		assert iface_name in self.bridges[bridge_name],'Interface doesn\'t exists'
		
		try:
			self.call_wrapper(['brctl','delif', bridge_name, iface_name])
			self.bridges[bridge_name].remove(iface_name)
			return True
		
		except:
			return False
			
	def get_bridge_members(self, name = None):
		cmd = ['brctl', 'show']
		if(name != None):
			cmd.append(name)
		
		dev_null = open(os.devnull, 'w')
		data = subprocess.check_output(cmd, stderr = dev_null)
		data = data.split('\n')[1:]
		bridge = None
		retVal = {}
	
		for line in data:
			if(line.strip() == ''):
				continue
			
			if(line[0] in ['', '\t']):
				if(bridge == None):
					continue
				
				else:
					retVal[bridge].append(line.strip())	
				
			else:
				line = line.split()
				bridge = line[0].strip()
				retVal[bridge] = []
				if(len(line) > 3):
					retVal[bridge].append(line[3])	
					
		if(name != None):
			return retVal[name]
		
		else:
			return retVal
	
	def setup_bridge(self, bridge_data):
		bridge_name = 'brcust{}'.format(bridge_data['id'])
		if(not self.iface_exists(bridge_name)):
			if(not self.create_bridge(bridge_name)):
				return [False, 'couldn\'t create bridge']
				
		self.active_ifaces.append(bridge_name)
		
		iface_error = []
		rest = list(self.bridges[bridge_name])
		for member_data in bridge_data['members']:
			res = self.setup_iface(member_data)
			if(not res[0]):
				iface_error.append(member_data)
				
			else:
				member_of = filter(lambda x: res[1] in self.bridges[x], self.bridges)
				if(len(member_of) == 0):
					if(not self.add_bridge_member(bridge_name, res[1])):
						iface_error.append('couldn\'t add {} to bridge {}'.format(res[1], bridge_name))
						
				elif(member_of[0] == bridge_name):
					rest.remove(res[1])
					continue
					
				else:
					iface_error.append('device {} is already a member of a bridge; can\'t enslave it to bridge {}'.format(res[1], bridge_name))
					
		for iface in rest:
			if(not self.remove_bridge_member(bridge_name, iface)):
				iface_error.append('couldn\'t remove {} from bridge {}'.format(iface, bridge_name))
				continue
				
			if(not self.remove_iface(iface)):
				iface_error.append('couldn\'t remove vlan interface {}'.format(iface))
				continue
					
			if(iface[:2] == 'vp'):
				vxlan = iface.split('.')[0]
				
				if(not self.remove_iface(vxlan)):
					iface_error.append('couldn\'t remove vxlan interface {}'.format(vxlan))
					continue
				
		if(len(iface_error)):
			return [False, iface_error]
			
		else:
			return [True, None]
			
	def setup_iface(self, iface_data):
		vxlan_iface = ''
		if(iface_data['public']):
			vxlan_iface = 'vp{}'.format(iface_data['vxlan_id'])
			if(not self.iface_exists(vxlan_iface)):
				if(self.create_vxlan(iface_data)):
					self.active_ifaces.append(vxlan_iface)
				
				else:
					return [False, 'couldn\'t create public vxlan interface']
					
		else:
			vxlan_iface = 'vv{}'.format(iface_data['vxlan_id'])
			if(not self.iface_exists(vxlan_iface)):
				return [False, 'private vxlan interface doesn\'t exist (not connected)']
			
			self.active_ifaces.append(vxlan_iface)	
				
		vlan_iface = vxlan_iface + '.{}'.format(iface_data['vlan_id'])
		if(not self.iface_exists(vlan_iface)):
			if(not self.create_vlan(iface_data)):
				return [False, 'couldn\'t create vlan interface']
		
		self.active_ifaces.append(vlan_iface)		
		return [True, vlan_iface]
	
	def apply_configuration(self):
		"""docstring for apply_configuration"""
		logger.info('applying configuration')
		if(self.bridge_data_list == None):
			return
			
		with self._lock:
			self.active_ifaces = []
			self.bridges = self.get_bridge_members()
			rest = filter(lambda x: x[:6] == 'brcust', self.bridges.keys())
			for bridge_data in self.bridge_data_list:
				res = self.setup_bridge(bridge_data)
				if('brcust{}'.format(bridge_data['id']) in rest):
					rest.remove('brcust{}'.format(bridge_data['id']))
					
				if(res[0] == False):
					map(lambda x: logger.error(x), res[1])
			
			for bridge_name in rest:
				members = list(self.bridges[bridge_name])
				for iface in members:
					if(not self.remove_bridge_member(bridge_name, iface)):
						iface_error.append('couldn\'t remove {} from bridge {}'.format(iface, bridge_name))
						continue
				
					if(not self.remove_iface(iface)):
						iface_error.append('couldn\'t remove vlan interface {}'.format(iface))
						continue
					
					if(iface[:2] == 'vp'):
						vxlan = iface.split('.')[0]
				
						if(not self.remove_iface(vxlan)):
							iface_error.append('couldn\'t remove vxlan interface {}'.format(vxlan))
							continue
				
				self.remove_bridge(bridge_name)
				
		logger.info('configuration applied')
		
	
	def signal(self,signum,frame):
		if signum == signal.SIGHUP:
			logger.info('received HUP signal - reloading')
			self.update_event.set()
			
		else:
			logger.info('received INT signal - terminating')
			self.stop_event.set()
	
	def run(self):
		"""docstring for main_loop"""
		logger.info('Running c2p bridge')
		while True:
			if self.stop_event.is_set():
				return
				
			self.fetch_bridge_data()
			if self.update_event.is_set():
				self.update_event.clear()
				self.apply_configuration()
				
			time.sleep(2)
